import { AnyObject } from '@loopback/repository';
import { RequestContext } from '@loopback/rest';
import { ILogger, SuccessResponse } from '@sourceloop/core';
import { RefreshTokenRequest } from '../models';
import { AuthTokenRequest, IdpConfiguration, TokenResponse } from '../modules/auth';
import { CodeReaderFn, JwtPayloadFn, JWTSignerFn, JWTVerifierFn } from '../providers';
import { AuthClientRepository, JwtKeysRepository, LoginActivityRepository, RefreshTokenRepository, RevokedTokenRepository, UserRepository, UserTenantRepository } from '../repositories';
import { ActorId, IUserActivity } from '../types';
export declare class IdpLoginService {
    authClientRepository: AuthClientRepository;
    userRepo: UserRepository;
    userTenantRepo: UserTenantRepository;
    refreshTokenRepo: RefreshTokenRepository;
    jwtKeysRepo: JwtKeysRepository;
    private readonly revokedTokensRepo;
    logger: ILogger;
    private readonly loginActivityRepo;
    private readonly actorKey;
    private readonly ctx;
    private readonly codeReader;
    private readonly jwtVerifier;
    private readonly jwtSigner;
    private readonly getJwtPayload;
    private readonly userActivity?;
    constructor(authClientRepository: AuthClientRepository, userRepo: UserRepository, userTenantRepo: UserTenantRepository, refreshTokenRepo: RefreshTokenRepository, jwtKeysRepo: JwtKeysRepository, revokedTokensRepo: RevokedTokenRepository, logger: ILogger, loginActivityRepo: LoginActivityRepository, actorKey: ActorId, ctx: RequestContext, codeReader: CodeReaderFn, jwtVerifier: JWTVerifierFn<AnyObject>, jwtSigner: JWTSignerFn<object>, getJwtPayload: JwtPayloadFn, userActivity?: IUserActivity | undefined);
    /**
     * The function `getOpenIdConfiguration` returns an IdpConfiguration object with specific properties
     * set based on environment variables.
     * @returns An IdpConfiguration object with the specified properties and values is being returned.
     */
    getOpenIdConfiguration(): Promise<IdpConfiguration>;
    /**
     * The function `generateToken` generates a JWT token for a client using a code
     * and performs various authentication checks.
     * @param {string} clientId - The `clientId` parameter in the `generateToken`
     * function is a string that represents the client ID associated with the
     * authentication request. It is used to identify the client making the request
     * for generating a token.
     * @param {string} code - The `code` parameter in the `generateToken` function is
     * a string that represents the authorization code that will be used to generate
     * a token for authentication and authorization purposes. This code is typically
     * obtained during the authorization code flow in OAuth 2.0 when a user grants
     * permission to a client application.
     * @param {CodeReaderFn} codeReader - The `codeReader` parameter in the
     * `generateToken` function is a function that reads a code and returns a result.
     * It is injected using `@inject(AuthCodeBindings.CODEREADER_PROVIDER)` which
     * means it is provided by a binding defined in the `AuthCodeBindings` namespace.
     * The
     * @returns The `generateToken` function is returning the result of calling
     * `this.createJWT(payload, authClient, LoginType.ACCESS)` after performing
     * various checks and operations.
     */
    generateToken(request: AuthTokenRequest): Promise<TokenResponse>;
    /**
     * The `createJWT` function generates a JWT token for a user with specified
     * payload and authentication client, handling user authentication and token
     * expiration.
     * @param payload - The `payload` parameter in the `createJWT` function is an
     * object that contains information about the user and external tokens. It has
     * the following properties:
     * @param {AuthClient} authClient - The `authClient` parameter in the `createJWT`
     * function represents the client that is requesting the JWT token generation. It
     * contains information about the client, such as the client ID, access token
     * expiration time, and refresh token expiration time. This information is used
     * to customize the JWT token generation process based
     * @param {LoginType} loginType - The `loginType` parameter in the `createJWT`
     * function represents the type of login being performed, such as regular login,
     * social login, or any other specific type of authentication method. It helps in
     * determining the context of the login operation and can be used to customize
     * the behavior or processing logic based
     * @param {string} [tenantId] - The `tenantId` parameter in the `createJWT`
     * function is an optional parameter of type string. It is used to specify the ID
     * of the tenant for which the JWT token is being created. If provided, it is
     * used in the process of generating the JWT payload. If not provided, the
     * @returns A `TokenResponse` object is being returned, which contains the
     * `accessToken`, `refreshToken`, and `expires` properties.
     */
    private createJWT;
    /**
     * The function `markUserActivity` encrypts and stores user login activity,
     * including IP address and payload, in a database.
     * @param {User} user - The `user` parameter in the `markUserActivity` function
     * represents the user who is performing the activity for which you are marking the
     * login activity. This user object likely contains information about the user,
     * such as their ID, name, email, etc. It is used to identify the actor of the
     * @param {UserTenant | null} userTenant - The `userTenant` parameter in the
     * `markUserActivity` function represents the tenant associated with the user. It
     * can be either an object of type `UserTenant` or `null` if there is no specific
     * tenant assigned to the user. The function uses this parameter to determine the
     * actor and tenant
     * @param {AnyObject} payload - The `payload` parameter in the `markUserActivity`
     * function is the data that you want to encrypt and store as part of the user's
     * login activity. In the provided code snippet, the `payload` is first converted
     * to a JSON string using `JSON.stringify(payload)`. Then, it is
     * @param {LoginType} loginType - The `loginType` parameter in the
     * `markUserActivity` function represents the type of login activity being
     * performed by the user. It is used to specify whether the user is logging in
     * using a certain method or platform. Examples of `loginType` could include
     * 'email', 'social', '2
     */
    private markUserActivity;
    /**
     * The `logoutUser` function in TypeScript handles the logout process for a user
     * by revoking tokens and deleting refresh tokens.
     * @param {string} auth - The `auth` parameter in the `logoutUser` function is a
     * string that represents the authentication token. It is used to identify and
     * authenticate the user who is attempting to log out. The function extracts the
     * token from the `auth` parameter and performs various checks and operations
     * related to user logout based on
     * @param {RefreshTokenRequest} req - The `req` parameter in the `logoutUser`
     * function is of type `RefreshTokenRequest`. It likely contains information
     * related to the refresh token that is used to identify and authenticate the
     * user during the logout process. This parameter may include properties such as
     * `refreshToken`, which is essential for revoking
     * @returns The `logoutUser` function returns a `Promise` that resolves to a
     * `SuccessResponse` object with a `success` property set to `true` and a `key`
     * property set to `refreshTokenModel.userId`.
     */
    logoutUser(auth: string, req: RefreshTokenRequest): Promise<SuccessResponse>;
    /**
     * The `generateKeys` function generates new keys for a specified number of times defined by the
     * `MAX_JWT_KEYS` environment variable or 2 by default.
     */
    generateKeys(): Promise<void>;
    /**
     * The function generates a JSON Web Key Set (JWKS) containing a RSA public key and saves it to a
     * file.
     */
    generateNewKey(): Promise<void>;
    /**
     * The function decodes a JWT token and returns the expiration time in milliseconds.
     * @param {string} token - The `token` parameter is a string that represents a JSON Web Token (JWT).
     * @returns the expiry time of the token in milliseconds.
     */
    /**
     * Decodes the given token and retrieves the expiry timestamp.
     *
     * @param token - The token to decode.
     * @returns The expiry timestamp in milliseconds.
     */
    decodeAndGetExpiry(token: string): number | null;
}
