"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.IdpLoginService = void 0;
const tslib_1 = require("tslib");
const core_1 = require("@loopback/core");
const repository_1 = require("@loopback/repository");
const rest_1 = require("@loopback/rest");
const core_2 = require("@sourceloop/core");
const crypto_1 = tslib_1.__importStar(require("crypto"));
const jwt = tslib_1.__importStar(require("jsonwebtoken"));
const moment_1 = tslib_1.__importDefault(require("moment"));
const jose = tslib_1.__importStar(require("node-jose"));
const keys_1 = require("../keys");
const models_1 = require("../models");
const auth_1 = require("../modules/auth");
const keys_2 = require("../providers/keys");
const repositories_1 = require("../repositories");
let IdpLoginService = class IdpLoginService {
    constructor(authClientRepository, userRepo, userTenantRepo, refreshTokenRepo, jwtKeysRepo, revokedTokensRepo, logger, loginActivityRepo, actorKey, ctx, codeReader, jwtVerifier, jwtSigner, getJwtPayload, userActivity) {
        this.authClientRepository = authClientRepository;
        this.userRepo = userRepo;
        this.userTenantRepo = userTenantRepo;
        this.refreshTokenRepo = refreshTokenRepo;
        this.jwtKeysRepo = jwtKeysRepo;
        this.revokedTokensRepo = revokedTokensRepo;
        this.logger = logger;
        this.loginActivityRepo = loginActivityRepo;
        this.actorKey = actorKey;
        this.ctx = ctx;
        this.codeReader = codeReader;
        this.jwtVerifier = jwtVerifier;
        this.jwtSigner = jwtSigner;
        this.getJwtPayload = getJwtPayload;
        this.userActivity = userActivity;
    }
    /**
     * The function `getOpenIdConfiguration` returns an IdpConfiguration object with specific properties
     * set based on environment variables.
     * @returns An IdpConfiguration object with the specified properties and values is being returned.
     */
    async getOpenIdConfiguration() {
        // sonarignore:start
        const config = new auth_1.IdpConfiguration();
        config.issuer = `${process.env.API_BASE_URL}`;
        config.authorization_endpoint = `${process.env.API_BASE_URL}/connect/auth`;
        config.token_endpoint = `${process.env.API_BASE_URL}/connect/token`;
        config.jwks_uri = `${process.env.API_BASE_URL}/connect/get-keys`;
        config.end_session_endpoint = `${process.env.API_BASE_URL}/connect/endsession`;
        config.response_types_supported = ['code', 'id_token', 'token'];
        config.scopes_supported = ['openid', 'email', 'phone', 'profile'];
        config.id_token_signing_alg_values_supported = [
            'RS256',
            'HS256',
            'RS384',
            'RS512',
        ];
        config.token_endpoint_auth_methods_supported = [
            'client_secret_basic',
            'client_secret_post',
        ];
        config.userinfo_endpoint = `${process.env.API_BASE_URL}/connect/userinfo`;
        return config;
        // sonarignore:end
    }
    /**
     * The function `generateToken` generates a JWT token for a client using a code
     * and performs various authentication checks.
     * @param {string} clientId - The `clientId` parameter in the `generateToken`
     * function is a string that represents the client ID associated with the
     * authentication request. It is used to identify the client making the request
     * for generating a token.
     * @param {string} code - The `code` parameter in the `generateToken` function is
     * a string that represents the authorization code that will be used to generate
     * a token for authentication and authorization purposes. This code is typically
     * obtained during the authorization code flow in OAuth 2.0 when a user grants
     * permission to a client application.
     * @param {CodeReaderFn} codeReader - The `codeReader` parameter in the
     * `generateToken` function is a function that reads a code and returns a result.
     * It is injected using `@inject(AuthCodeBindings.CODEREADER_PROVIDER)` which
     * means it is provided by a binding defined in the `AuthCodeBindings` namespace.
     * The
     * @returns The `generateToken` function is returning the result of calling
     * `this.createJWT(payload, authClient, LoginType.ACCESS)` after performing
     * various checks and operations.
     */
    async generateToken(request) {
        const authClient = await this.authClientRepository.findOne({
            where: {
                clientId: request.clientId,
            },
        });
        if (!authClient) {
            throw new rest_1.HttpErrors.Unauthorized("Client Invalid" /* AuthErrorKeys.ClientInvalid */);
        }
        try {
            const resultCode = await this.codeReader(request.code);
            const payload = (await this.jwtVerifier(resultCode, {
                audience: request.clientId,
            }));
            if (payload.mfa) {
                throw new rest_1.HttpErrors.Unauthorized("User Verification Failed" /* AuthErrorKeys.UserVerificationFailed */);
            }
            if (payload.userId &&
                !(await this.userRepo.firstTimeUser(payload.userId))) {
                await this.userRepo.updateLastLogin(payload.userId);
            }
            return await this.createJWT(payload, authClient, "ACCESS" /* LoginType.ACCESS */);
        }
        catch (error) {
            this.logger.error(error);
            if (error.name === 'TokenExpiredError') {
                throw new rest_1.HttpErrors.Unauthorized("Code Expired" /* AuthErrorKeys.CodeExpired */);
            }
            else if (rest_1.HttpErrors.HttpError.prototype.isPrototypeOf(error)) {
                throw error;
            }
            else {
                throw new rest_1.HttpErrors.Unauthorized("Invalid Credentials" /* AuthErrorKeys.InvalidCredentials */);
            }
        }
    }
    /**
     * The `createJWT` function generates a JWT token for a user with specified
     * payload and authentication client, handling user authentication and token
     * expiration.
     * @param payload - The `payload` parameter in the `createJWT` function is an
     * object that contains information about the user and external tokens. It has
     * the following properties:
     * @param {AuthClient} authClient - The `authClient` parameter in the `createJWT`
     * function represents the client that is requesting the JWT token generation. It
     * contains information about the client, such as the client ID, access token
     * expiration time, and refresh token expiration time. This information is used
     * to customize the JWT token generation process based
     * @param {LoginType} loginType - The `loginType` parameter in the `createJWT`
     * function represents the type of login being performed, such as regular login,
     * social login, or any other specific type of authentication method. It helps in
     * determining the context of the login operation and can be used to customize
     * the behavior or processing logic based
     * @param {string} [tenantId] - The `tenantId` parameter in the `createJWT`
     * function is an optional parameter of type string. It is used to specify the ID
     * of the tenant for which the JWT token is being created. If provided, it is
     * used in the process of generating the JWT payload. If not provided, the
     * @returns A `TokenResponse` object is being returned, which contains the
     * `accessToken`, `refreshToken`, and `expires` properties.
     */
    async createJWT(payload, authClient, loginType, tenantId) {
        var _a;
        try {
            const size = 32;
            const ms = 1000;
            let user;
            if (payload.user) {
                user = payload.user;
            }
            else if (payload.userId) {
                user = await this.userRepo.findById(payload.userId, {
                    include: [
                        {
                            relation: 'defaultTenant',
                        },
                    ],
                });
                if (payload.externalAuthToken && payload.externalRefreshToken) {
                    user.externalAuthToken = payload.externalAuthToken;
                    user.externalRefreshToken =
                        payload.externalRefreshToken;
                }
            }
            else {
                // Do nothing and move ahead
            }
            if (!user) {
                throw new rest_1.HttpErrors.Unauthorized("UserDoesNotExist" /* AuthenticateErrorKeys.UserDoesNotExist */);
            }
            const data = await this.getJwtPayload(user, authClient, tenantId);
            const accessToken = await this.jwtSigner(data, {
                expiresIn: authClient.accessTokenExpiration,
            });
            const refreshToken = crypto_1.default.randomBytes(size).toString('hex');
            // Set refresh token into redis for later verification
            await this.refreshTokenRepo.set(refreshToken, {
                clientId: authClient.clientId,
                userId: user.id,
                username: user.username,
                accessToken,
                externalAuthToken: user.externalAuthToken,
                externalRefreshToken: user.externalRefreshToken,
                tenantId: data.tenantId,
            }, { ttl: authClient.refreshTokenExpiration * ms });
            const userTenant = await this.userTenantRepo.findOne({
                where: { userId: user.id },
            });
            if ((_a = this.userActivity) === null || _a === void 0 ? void 0 : _a.markUserActivity)
                this.markUserActivity({ ...data }, user, userTenant, loginType);
            return new auth_1.TokenResponse({
                accessToken,
                refreshToken,
                expires: (0, moment_1.default)()
                    .add(authClient.accessTokenExpiration, 's')
                    .toDate()
                    .getTime(),
            });
        }
        catch (error) {
            this.logger.error(error);
            if (rest_1.HttpErrors.HttpError.prototype.isPrototypeOf(error)) {
                throw error;
            }
            else {
                throw new rest_1.HttpErrors.Unauthorized("Invalid Credentials" /* AuthErrorKeys.InvalidCredentials */);
            }
        }
    }
    /**
     * The function `markUserActivity` encrypts and stores user login activity,
     * including IP address and payload, in a database.
     * @param {User} user - The `user` parameter in the `markUserActivity` function
     * represents the user who is performing the activity for which you are marking the
     * login activity. This user object likely contains information about the user,
     * such as their ID, name, email, etc. It is used to identify the actor of the
     * @param {UserTenant | null} userTenant - The `userTenant` parameter in the
     * `markUserActivity` function represents the tenant associated with the user. It
     * can be either an object of type `UserTenant` or `null` if there is no specific
     * tenant assigned to the user. The function uses this parameter to determine the
     * actor and tenant
     * @param {AnyObject} payload - The `payload` parameter in the `markUserActivity`
     * function is the data that you want to encrypt and store as part of the user's
     * login activity. In the provided code snippet, the `payload` is first converted
     * to a JSON string using `JSON.stringify(payload)`. Then, it is
     * @param {LoginType} loginType - The `loginType` parameter in the
     * `markUserActivity` function represents the type of login activity being
     * performed by the user. It is used to specify whether the user is logging in
     * using a certain method or platform. Examples of `loginType` could include
     * 'email', 'social', '2
     */
    markUserActivity(payload, user, userTenant, loginType) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k;
        const size = 16;
        const encryptionKey = process.env.ENCRYPTION_KEY;
        if (encryptionKey) {
            const iv = crypto_1.default.randomBytes(size);
            /* encryption of IP Address */
            const cipherIp = crypto_1.default.createCipheriv('aes-256-gcm', encryptionKey, iv);
            const ip = (_d = (_b = (_a = this.ctx.request.headers['x-forwarded-for']) === null || _a === void 0 ? void 0 : _a.toString()) !== null && _b !== void 0 ? _b : (_c = this.ctx.request.socket.remoteAddress) === null || _c === void 0 ? void 0 : _c.toString()) !== null && _d !== void 0 ? _d : '';
            const encyptIp = Buffer.concat([
                cipherIp.update(ip, 'utf8'),
                cipherIp.final(),
            ]);
            const authTagIp = cipherIp.getAuthTag();
            const ipAddress = JSON.stringify({
                iv: iv.toString('hex'),
                encryptedData: encyptIp.toString('hex'),
                authTag: authTagIp.toString('hex'),
            });
            /* encryption of Paylolad Address */
            const cipherPayload = crypto_1.default.createCipheriv('aes-256-gcm', encryptionKey, iv);
            let activityPayload;
            if (loginType === "LOGOUT" /* LoginType.LOGOUT */) {
                activityPayload = JSON.stringify({
                    ...user,
                    clientId: payload.clientId,
                });
            }
            else {
                activityPayload = JSON.stringify(payload);
            }
            const encyptPayload = Buffer.concat([
                cipherPayload.update(activityPayload, 'utf8'),
                cipherPayload.final(),
            ]);
            const authTagPayload = cipherIp.getAuthTag();
            const tokenPayload = JSON.stringify({
                iv: iv.toString('hex'),
                encryptedData: encyptPayload.toString('hex'),
                authTag: authTagPayload.toString('hex'),
            });
            // make an entry to mark the users login activity
            let actor, tenantId;
            if (userTenant) {
                actor = (_f = (_e = userTenant[this.actorKey]) === null || _e === void 0 ? void 0 : _e.toString()) !== null && _f !== void 0 ? _f : '0';
                tenantId = userTenant.tenantId;
            }
            else {
                actor = (_h = (_g = user['id']) === null || _g === void 0 ? void 0 : _g.toString()) !== null && _h !== void 0 ? _h : '0';
                tenantId = (_j = user.defaultTenantId) !== null && _j !== void 0 ? _j : '0';
            }
            const loginActivity = new models_1.LoginActivity({
                actor,
                tenantId,
                loginTime: new Date(),
                tokenPayload,
                loginType,
                deviceInfo: (_k = this.ctx.request.headers['user-agent']) === null || _k === void 0 ? void 0 : _k.toString(),
                ipAddress,
            });
            this.loginActivityRepo.create(loginActivity).catch(() => {
                this.logger.error(`Failed to add the login activity => ${JSON.stringify(loginActivity)}`);
            });
        }
    }
    /**
     * The `logoutUser` function in TypeScript handles the logout process for a user
     * by revoking tokens and deleting refresh tokens.
     * @param {string} auth - The `auth` parameter in the `logoutUser` function is a
     * string that represents the authentication token. It is used to identify and
     * authenticate the user who is attempting to log out. The function extracts the
     * token from the `auth` parameter and performs various checks and operations
     * related to user logout based on
     * @param {RefreshTokenRequest} req - The `req` parameter in the `logoutUser`
     * function is of type `RefreshTokenRequest`. It likely contains information
     * related to the refresh token that is used to identify and authenticate the
     * user during the logout process. This parameter may include properties such as
     * `refreshToken`, which is essential for revoking
     * @returns The `logoutUser` function returns a `Promise` that resolves to a
     * `SuccessResponse` object with a `success` property set to `true` and a `key`
     * property set to `refreshTokenModel.userId`.
     */
    async logoutUser(auth, req) {
        var _a;
        const token = auth === null || auth === void 0 ? void 0 : auth.replace(/bearer /i, '');
        if (!token || !req.refreshToken) {
            throw new rest_1.HttpErrors.UnprocessableEntity("TokenMissing" /* AuthenticateErrorKeys.TokenMissing */);
        }
        const refreshTokenModel = await this.refreshTokenRepo.get(req.refreshToken);
        if (!refreshTokenModel) {
            throw new rest_1.HttpErrors.Unauthorized("Token Expired" /* AuthErrorKeys.TokenExpired */);
        }
        if (refreshTokenModel.accessToken !== token) {
            throw new rest_1.HttpErrors.Unauthorized("Token Invalid" /* AuthErrorKeys.TokenInvalid */);
        }
        const expiry = this.decodeAndGetExpiry(token);
        await this.revokedTokensRepo.set(token, { token }, {
            ttl: expiry,
        });
        await this.revokedTokensRepo.set(token, { token });
        await this.refreshTokenRepo.delete(req.refreshToken);
        if (refreshTokenModel.pubnubToken) {
            await this.refreshTokenRepo.delete(refreshTokenModel.pubnubToken);
        }
        const user = await this.userRepo.findById(refreshTokenModel.userId);
        const userTenant = await this.userTenantRepo.findOne({
            where: { userId: user.id },
        });
        if ((_a = this.userActivity) === null || _a === void 0 ? void 0 : _a.markUserActivity)
            this.markUserActivity(refreshTokenModel, user, userTenant, "LOGOUT" /* LoginType.LOGOUT */);
        return new core_2.SuccessResponse({
            success: true,
            key: refreshTokenModel.userId,
        });
    }
    /**
     * The `generateKeys` function generates new keys for a specified number of times defined by the
     * `MAX_JWT_KEYS` environment variable or 2 by default.
     */
    async generateKeys() {
        var _a;
        //call the function to generate new keys for process.env.MAX_JWT_KEYS times.
        for (let i = 0; i < +((_a = process.env.MAX_JWT_KEYS) !== null && _a !== void 0 ? _a : 2); i++) {
            await this.generateNewKey();
        }
    }
    /**
     * The function generates a JSON Web Key Set (JWKS) containing a RSA public key and saves it to a
     * file.
     */
    async generateNewKey() {
        var _a;
        // Generate the RSA key pair
        const { publicKey, privateKey } = (0, crypto_1.generateKeyPairSync)('rsa', {
            modulusLength: 2048,
            publicKeyEncoding: {
                type: 'spki',
                format: 'pem',
            },
            privateKeyEncoding: {
                type: 'pkcs8',
                format: 'pem',
                cipher: 'aes-256-cbc',
                passphrase: process.env.JWT_PRIVATE_KEY_PASSPHRASE,
            },
        });
        // Create the JWKS object
        const keyStore = jose.JWK.createKeyStore();
        const key = await keyStore.add(publicKey, 'pem');
        // Fetch existing keys from the database
        const existingKeys = await this.jwtKeysRepo.find({
            order: ['createdOn ASC'],
        });
        // Rotate keys: remove the oldest key if max limit is exceeded
        const maxKeys = +((_a = process.env.MAX_JWT_KEYS) !== null && _a !== void 0 ? _a : 2);
        if (existingKeys.length >= maxKeys) {
            const oldestKey = existingKeys[0];
            await this.jwtKeysRepo.deleteById(oldestKey.id);
        }
        // Save public and private keys to the database using JwtKeysRepository
        await this.jwtKeysRepo.create({
            keyId: key.kid,
            publicKey: publicKey,
            privateKey: privateKey,
        });
    }
    /**
     * The function decodes a JWT token and returns the expiration time in milliseconds.
     * @param {string} token - The `token` parameter is a string that represents a JSON Web Token (JWT).
     * @returns the expiry time of the token in milliseconds.
     */
    /**
     * Decodes the given token and retrieves the expiry timestamp.
     *
     * @param token - The token to decode.
     * @returns The expiry timestamp in milliseconds.
     */
    decodeAndGetExpiry(token) {
        const tokenData = jwt.decode(token); // handle null result from decode
        const ms = 1000;
        if (tokenData === null || tokenData === void 0 ? void 0 : tokenData.exp) {
            return tokenData.exp * ms;
        }
        // If tokenData or exp is missing, return null to indicate no expiry
        return null;
    }
};
exports.IdpLoginService = IdpLoginService;
exports.IdpLoginService = IdpLoginService = tslib_1.__decorate([
    (0, core_1.injectable)({ scope: core_1.BindingScope.TRANSIENT }),
    tslib_1.__param(0, (0, repository_1.repository)(repositories_1.AuthClientRepository)),
    tslib_1.__param(1, (0, repository_1.repository)(repositories_1.UserRepository)),
    tslib_1.__param(2, (0, repository_1.repository)(repositories_1.UserTenantRepository)),
    tslib_1.__param(3, (0, repository_1.repository)(repositories_1.RefreshTokenRepository)),
    tslib_1.__param(4, (0, repository_1.repository)(repositories_1.JwtKeysRepository)),
    tslib_1.__param(5, (0, repository_1.repository)(repositories_1.RevokedTokenRepository)),
    tslib_1.__param(6, (0, core_1.inject)(core_2.LOGGER.LOGGER_INJECT)),
    tslib_1.__param(7, (0, repository_1.repository)(repositories_1.LoginActivityRepository)),
    tslib_1.__param(8, (0, core_1.inject)(keys_1.AuthServiceBindings.ActorIdKey)),
    tslib_1.__param(9, core_1.inject.context()),
    tslib_1.__param(10, (0, core_1.inject)(keys_2.AuthCodeBindings.CODEREADER_PROVIDER)),
    tslib_1.__param(11, (0, core_1.inject)(keys_2.AuthCodeBindings.JWT_VERIFIER, { optional: true })),
    tslib_1.__param(12, (0, core_1.inject)(keys_2.AuthCodeBindings.JWT_SIGNER)),
    tslib_1.__param(13, (0, core_1.inject)(keys_1.AuthServiceBindings.JWTPayloadProvider)),
    tslib_1.__param(14, (0, core_1.inject)(keys_1.AuthServiceBindings.MarkUserActivity, { optional: true })),
    tslib_1.__metadata("design:paramtypes", [repositories_1.AuthClientRepository,
        repositories_1.UserRepository,
        repositories_1.UserTenantRepository,
        repositories_1.RefreshTokenRepository,
        repositories_1.JwtKeysRepository,
        repositories_1.RevokedTokenRepository, Object, repositories_1.LoginActivityRepository, String, rest_1.RequestContext, Function, Function, Function, Function, Object])
], IdpLoginService);
//# sourceMappingURL=idp-login.service.js.map